###############################################################################
rm(list=ls())

# load some packages
library(rstan)
library(deSolve)
library(coda)
library(vioplot)
library(loo)

###############################################################################
# load data
data_para <- read.table(file="growth_travelling.txt", header=T)

###############################################################################
# data structure to save posteriors
all_r0_dist_list_rK <- list()
all_K_dist_list_rK <- list()
all_N0sim_dist_list_rK <- list()

all_r0_dist_list_BH <- list()
all_K_dist_list_BH <- list()
all_d_dist_list_BH <- list()
all_N0sim_dist_list_BH <- list()

all_tubeID_list <- list()

all_waic_rK <- numeric()
all_waic_BH <- numeric()
all_waic_tubeID_vect <- numeric()

###############################################################################
# declare stan models

# stan rK model
stanmodelcode_rK = '
// function that caclulates r-alpha logistic population growth
functions{
  real[] odemodel(real t, real[] N, real[] p, real[] x_r, int[] x_i){
    // p[1]=r0, p[2]=K
    real dNdt[1];
    dNdt[1] = p[1] * (1 - (N[1]/p[2])) * N[1];
  return dNdt;
  }
}

data{
  int n;
  real log_N0;
  real log_N[n];
  real t0;
  real t[n];
}

transformed data { 
  // not used here
  real x_r[0];
  int x_i[0];
}

parameters{
  real log_r;
  real log_K;
  real log_N0sim;
  real<lower=0> sdev;
}

transformed parameters{
  // all of this was in the model section previously
  // I moved it here to be able to get waic because Nsim needs to be accessible in "generated quantities {}"
  
  real p[2];
  real Nsim[n,1]; // simulated values, matrix. dim1 = time, dim2 = dim_ODE = 1
  real N0sim_dummy[1]; // just a dummy because the ODE solver requires real[] instead of real
  
  // parameters for integrator
  p[1] = exp(log_r);
  p[2] = exp(log_K);
  N0sim_dummy[1] = exp(log_N0sim);  // see above
  
  // integrate ODE
  Nsim = integrate_ode_rk45(odemodel,N0sim_dummy,t0,t,p,x_r,x_i);
}

model{
  // priors
  // note: it can be VERY helpful to estimate parameters on logscale, 
  // especially if they have different orders of magnitude. here it works on regular scale, though.
  log_r ~ normal(log(0.1), 1.5);
  log_K ~ normal(log(500), 1);
  log_N0sim ~ normal(log(2),1);
  sdev ~ cauchy(0,1);
  
  // likelihood, normal (maybe lognormal helpful)
  log_N0 ~ normal(log_N0sim,sdev);
  for (i in 1:n){
    log_N[i] ~ normal(log(Nsim[i,1]),sdev);
  }  
}

// cacluate log lik to get waic
// from loo R package description
generated quantities {
  real log_lik[n];
  for (nn in 1:n){ 
  log_lik[nn] = normal_lpdf(log_N[nn] | log(Nsim[nn,1]), sdev);
  }
}
'

# stan BH model

# stan BH model wokring with înput of priors
stanmodelcode_BH = '
  // function that caclulates BH population growth
functions{
  real[] odemodel(real t, real[] N, real[] p, real[] x_r, int[] x_i){
    // p[1]=r0, p[2]=d, p[3]=K
    real dNdt[1];
    dNdt[1] = ((p[1] + p[2])/(1 + ((p[1]/(p[3]*p[2])) * N[1])) - p[2])*N[1];
    return dNdt;
  }
}

data{
  int n;
  real log_N0;
  real log_N[n];
  real t0;
  real t[n];
}

transformed data { 
  // not used here
  real x_r[0];
  int x_i[0];
}

parameters{
  real log_r;
  real log_d;
  real log_K;
  real log_N0sim;
  real<lower=0> sdev;
}


transformed parameters{
  // all of this was in the model section previously
  // I moved it here to be able to get waic because Nsim needs to be accessible in "generated quantities {}"
  
  real p[3];
  real Nsim[n,1]; // simulated values, matrix. dim1 = time, dim2 = dim_ODE = 1
  real N0sim_dummy[1]; // just a dummy because the ODE solver requires real[] instead of real
  
  // parameters for integrator
  p[1] = exp(log_r);
  p[2] = exp(log_d);
  p[3] = exp(log_K);
  N0sim_dummy[1] = exp(log_N0sim);  // see above
  
  // integrate ODE (maybe try: integrate_ode_bdf()??)
  Nsim = integrate_ode_rk45(odemodel,N0sim_dummy,t0,t,p,x_r,x_i);
}

model{
  // priors
  // note: it can be VERY helpful to estimate parameters on logscale, 
  // especially if they have different orders of magnitude. here it works on regular scale, though.
  log_r ~ normal(log(0.1), 1.5);
  log_K ~ normal(log(500), 1);
  log_d ~ normal(log(0.1),1.5);
  log_N0sim ~ normal(log(2),1);
  sdev ~ cauchy(0,1);
  
  // likelihood, normal (maybe lognormal helpful)
  log_N0 ~ normal(log_N0sim,sdev);
  for (i in 1:n){
    log_N[i] ~ normal(log(Nsim[i,1]),sdev);
  }  
}

// cacluate log lik to get waic
// from loo R package description
generated quantities {
  real log_lik[n];
  for (nn in 1:n){ 
    log_lik[nn] = normal_lpdf(log_N[nn] | log(Nsim[nn,1]), sdev);
  }
}
'

###############################################################################
# functions
ode.model_BH = function(t,N,p){
  with(as.list(p),{
    dNdt = ((p[1] + p[2])/(1 + ((p[1]/(p[3]*p[2])) * N)) - p[2])*N
    return(list(dNdt))
  })
}

ode.model_rK = function(t,N,p){
  with(as.list(p),{
    dNdt = p[1]*(1 - N/p[2])*N
    return(list(dNdt))
  })
}

###############################################################################
# stan options
chains = 1
rstan_options(auto_write = TRUE)
options(mc.cores = chains) 
iter   =  10000
warmup =  1000
thin   =     1

# compile models
s_model_rK = stan_model(model_code=stanmodelcode_rK)
s_model_BH = stan_model(model_code=stanmodelcode_BH)

###############################################################################
# general plot settings
#x11(width=7.8,height=9.1)
#par(mfrow=c(7,6),bty="l",mar=rep(0.5,4),oma=c(4,4,0.5,0.5))

###############################################################################
# global counter var
cnt <- 0

# loop over tubes
for(act_tubeID in unique(data_para$tube_ID)){
  cnt <- cnt + 1
  
  ########################################
  # get the subset
  act_subset <- which(data_para$tube_ID==act_tubeID)
  
  # get the densities and current time points
  # n0 was not measured, but is known to be 20 indivs in 15mL
  # this is why we add t0 = 0 to the time vect
  act_densities <- c(20/15,data_para$inds_mL[act_subset])
  act_times <- c(0,data_para$time[act_subset] * 24)
  
  #plot(act_densities ~ act_times,type="b")
  
  # check whether the time series starts with a zero
  if (sum(act_densities==0)!=0){
     act_times <- act_times[which(act_densities!=0)]
     act_densities <- act_densities[which(act_densities!=0)]
   }
  
  ###############################################################################
  # rK fitting
  
  # create data object for rstan
  data_rK = list(n = length(act_times)-1,
                 log_N0 = log(act_densities[1]),
                 log_N  = log(act_densities[2:length(act_times)]),
                 t0    = act_times[1],
                 t     = act_times[2:length(act_times)])
  
  # initial values for rK fitting
  init_rK=rep(list(list(log_r=log(0.1),
                        log_K = log(500),
                        log_N0sim=log(2),
                        sdev=1))
              ,chains)
  
  # do the rK fit
  fit_rK = sampling(s_model_rK,
                    data=data_rK,
                    iter=iter,
                    warmup=warmup,
                    thin=thin,
                    chains=chains,
                    init=init_rK
  )
  
  # check model
  print(fit_rK)
  
  # convert to mcmc object and plot posterior distributions
  samples_rK <- mcmc.list(lapply(1:ncol(fit_rK), function(x) mcmc(as.array(fit_rK)[,1,][,names(init_rK[[1]])])))
  #plot(samples_rK)
  
  ###############################################################################
  # BH fitting
  
  # create data object for rstan
  data_BH = list(n = length(act_times)-1,
                 log_N0 = log(act_densities[1]),
                 log_N  = log(act_densities[2:length(act_times)]),
                 t0    = act_times[1],
                 t     = act_times[2:length(act_times)])
  
  # initial values for rK fitting
  init_BH=rep(list(list(log_r=log(0.1),
                        log_K = log(500),
                        log_d = log(0.05),
                        log_N0sim=log(2),
                        sdev=1))
              ,chains)
  
  # do the rK fit
  fit_BH = sampling(s_model_BH,
                    data=data_BH,
                    iter=iter,
                    warmup=warmup,
                    thin=thin,
                    chains=chains,
                    init=init_BH
  )
  
  # check model
  print(fit_BH)
  
  # convert to mcmc object and plot posterior distributions
  samples_BH <- mcmc.list(lapply(1:ncol(fit_BH), function(x) mcmc(as.array(fit_BH)[,1,][,names(init_BH[[1]])])))
  #plot(samples_BH)
  
  ###############################################################################
  # compare fits
  all_waic_rK[cnt] <- waic(extract_log_lik(fit_rK))$waic
  all_waic_BH[cnt] <- waic(extract_log_lik(fit_BH))$waic
  all_waic_tubeID_vect[cnt] <- act_tubeID
  
  ###############################################################################
  # save posterios distributions of parameters of interest
  all_r0_dist_list_rK[[cnt]] <- exp(as.numeric(samples_rK[[1]][,"log_r"]))
  all_K_dist_list_rK[[cnt]] <- exp(as.numeric(samples_rK[[1]][,"log_K"]))
  all_N0sim_dist_list_rK[[cnt]] <- exp(as.numeric(samples_rK[[1]][,"log_N0sim"]))
  
  all_r0_dist_list_BH[[cnt]] <- exp(as.numeric(samples_BH[[1]][,"log_r"]))
  all_K_dist_list_BH[[cnt]] <- exp(as.numeric(samples_BH[[1]][,"log_K"]))
  all_d_dist_list_BH[[cnt]] <- exp(as.numeric(samples_BH[[1]][,"log_d"]))
  all_N0sim_dist_list_BH[[cnt]] <- exp(as.numeric(samples_BH[[1]][,"log_N0sim"]))
  
  all_tubeID_list[[cnt]] <- rep(act_tubeID,length(all_r0_dist_list_rK[[cnt]]))
  
  ###############################################################################
  # get predictions
  # set time interval for prediction
  timessim <- seq(min(act_times),max(act_times),len=100)
  
  # counter for color
  all_cols <- c("black","blue")
  act_col_cnt <- 0
  
  # get median model
  Nsim_rK = as.data.frame(lsoda(y=median(exp(as.numeric(samples_rK[[1]][,"log_N0sim"]))),
                                times=timessim,
                                func=ode.model_rK,
                                parms=c(median(exp(as.numeric(samples_rK[[1]][,"log_r"]))), median(exp(as.numeric(samples_rK[[1]][,"log_K"]))))))[,2]
  
  # help function for prediction of CI
  hlp_pred <- function(act_parms_no){
    as.data.frame(lsoda(y=exp(as.numeric(samples_rK[[1]][act_parms_no,"log_N0sim"])),
                        times=timessim,
                        func=ode.model_rK,
                        parms=exp(as.numeric(samples_rK[[1]][act_parms_no,c("log_r","log_K")]))))[,2]
  }
  
  # predict distribution at every time point
  pred_pop_dyn <- sapply(1:dim(samples_rK[[1]])[1],hlp_pred)
  
  # function for extracting CI and median of prediction
  get_quantile <- function(act_t_no){
    quantile(pred_pop_dyn[act_t_no,],p=c(0.025,0.5,0.975),na.rm=T)
  }
  
  # get prediction
  model_predictions_rK <- sapply(1:length(timessim),get_quantile)
  
  
  ##
  # get median model
  Nsim_BH = as.data.frame(lsoda(y=median(exp(as.numeric(samples_BH[[1]][,"log_N0sim"]))),
                                times=timessim,
                                func=ode.model_BH,
                                parms=c(median(exp(as.numeric(samples_BH[[1]][,"log_r"]))), median(exp(as.numeric(samples_BH[[1]][,"log_d"]))), median(exp(as.numeric(samples_BH[[1]][,"log_K"]))))))[,2]
  
  # help function for prediction of CI
  hlp_pred <- function(act_parms_no){
    as.data.frame(lsoda(y=exp(as.numeric(samples_BH[[1]][act_parms_no,"log_N0sim"])),
                        times=timessim,
                        func=ode.model_BH,
                        parms=exp(as.numeric(samples_BH[[1]][act_parms_no,c("log_r","log_d","log_K")]))))[,2]
  }
  
  # predict distribution at every time point
  pred_pop_dyn <- sapply(1:dim(samples_BH[[1]])[1],hlp_pred)
  
  # function for extracting CI and median of prediction
  get_quantile <- function(act_t_no){
    quantile(pred_pop_dyn[act_t_no,],p=c(0.025,0.5,0.975),na.rm=T)
  }
  
  # get prediction
  model_predictions_BH <- sapply(1:length(timessim),get_quantile)
  
  ###############################################################################
  # plot current growth dynamics
  
  par(bty="l",mar=c(4.5,4.5,1,1))
  plot(act_densities ~ act_times, pch=16, type="n", lty=3, ylab=expression(paste("Population density, N (indiv. ",mL^-1,")")), xlab="Time, t (h)", ylim=c(0,1000), xlim=c(0,max(data_para$time)*24) ,col="darkblue")
  points(act_densities ~ act_times, pch=16, type="b", lty=3, col="black",cex=1.5)
  
  act_col_cnt <- 1
  polygon(x = c(timessim,rev(timessim)), y= c(model_predictions_rK[1,], rev(model_predictions_rK[3,])),border=F,col=adjustcolor(all_cols[act_col_cnt],alpha.f=0.2))
  lines(Nsim_rK~timessim,col=adjustcolor(all_cols[act_col_cnt],alpha.f=0.2),lwd=2)
  
  act_col_cnt <- 2
  polygon(x = c(timessim,rev(timessim)), y= c(model_predictions_BH[1,], rev(model_predictions_BH[3,])),border=F,col=adjustcolor(all_cols[act_col_cnt],alpha.f=0.2))
  lines(Nsim_BH~timessim,col=adjustcolor(all_cols[act_col_cnt],alpha.f=0.2),lwd=2)
  
  legend("bottomright",fill=c("grey","darkblue"),legend=c("logistic model","Beverton-Holt model"),bty="n")
  
  ###############################################################################
  # save plot
  dev.copy2pdf(file=paste0("Para_dynamics_fit_",act_tubeID,".pdf"))
}

###############################################################################

write.table(
  cbind(
    unlist(all_r0_dist_list_BH),
    unlist(all_K_dist_list_BH),
    unlist(all_d_dist_list_BH),
    unlist(all_tubeID_list)
  ),
  "Para_dynamics_BH_post.txt",quote = F, row.names = F,
  col.names = c("r0","Nhat","d","tube_ID")
)

write.table(
  cbind(
    unlist(all_r0_dist_list_rK),
    unlist(all_K_dist_list_rK),
    unlist(all_tubeID_list)
  ),
  "Para_dynamics_rK_post.txt",quote = F, row.names = F,
  col.names = c("r0","Nhat","tube_ID")
)

write.table(
  cbind(
    all_waic_rK,
    all_waic_BH,
    all_waic_tubeID_vect
  ),
  "Para_dynamics_all_waic.txt",quote = F, row.names = F,
  col.names = c("waic_rK","waic_BH","tube_ID")
)

###############################################################################